/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <cstdint>

namespace glaxnimate::io {

using VarUint = std::uint64_t;
using Float32 = float;

} // namespace glaxnimate::io
