/*
 * SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#define PROJECT_VERSION "7.37.0"
#define PROJECT_SLUG ""
#define PROJECT_NAME "MLT"
#define URL_DOCS ""
#define URL_ISSUES ""
#define URL_DONATE ""
#define PROJECT_DESCRIPTION "Multimedia Framework"
