/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "glaxnimate/model/shapes/shapes/path.hpp"

GLAXNIMATE_OBJECT_IMPL(glaxnimate::model::Path)

