#pragma once
#ifndef _ofxOpenGLRender_h_
#define _ofxOpenGLRender_h_

// Copyright OpenFX and Contributors to the OpenFX project.
// SPDX-License-Identifier: BSD-3-Clause

#endif  // _ofxOpenGLRender_h_
