/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <array>
#include <QPointF>

namespace glaxnimate::math::bezier {


using BezierSegment = std::array<QPointF, 4>;


} // namespace glaxnimate::math::bezier
