/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include "glaxnimate/io/base.hpp"
#include "glaxnimate/io/io_registry.hpp"

namespace glaxnimate::io::avd {


class AvdFormat : public ImportExport
{
    Q_OBJECT

public:
    QString slug() const override { return "avd"; }
    QString name() const override { return i18n("Android Vector Drawable"); }
    QStringList extensions(Direction) const override { return {"xml"}; }
    bool can_save() const override { return true; }
    bool can_open() const override { return true; }

protected:
    bool on_open(QIODevice& file, const QString&, model::Document* document, const QVariantMap& options) override;
    bool on_save(QIODevice & file, const QString & filename, model::Composition* comp, const QVariantMap & setting_values) override;
};


} // namespace glaxnimate::io::avd


