// The only purpose of this file is to DEFINE the catch config so it can include main()

#define CATCH_CONFIG_MAIN // This tells Catch to provide a main() - only do this in one cpp file

#include "catch.hpp"
