/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
#define THORVG_SW_RASTER_SUPPORT 1

#cmakedefine THORVG_THREAD_SUPPORT 1

#cmakedefine THORVG_GL_RASTER_SUPPORT 1
#cmakedefine THORVG_GL_TARGET_GL 1
#cmakedefine THORVG_GL_TARGET_GLES 1

// #define THORVG_WG_RASTER_SUPPORT 0
// #define THORVG_PARTIAL_RENDER_SUPPORT 0
// #define THORVG_SVG_LOADER_SUPPORT 0
// #define THORVG_PNG_LOADER_SUPPORT 0
// #define THORVG_JPG_LOADER_SUPPORT 0
// #define THORVG_LOTTIE_LOADER_SUPPORT 0
// #define THORVG_TTF_LOADER_SUPPORT 0
// #define THORVG_WEBP_LOADER_SUPPORT 0
// #define THORVG_GIF_SAVER_SUPPORT 0
// #define THORVG_AVX_VECTOR_SUPPORT 0
// #define THORVG_NEON_VECTOR_SUPPORT 0
// #define THORVG_CAPI_BINDING_SUPPORT 0
// #define THORVG_LOG_ENABLED 0
// #define THORVG_FILE_IO_SUPPORT 0
// #define THORVG_LOTTIE_EXPRESSIONS_SUPPORT 0


#define THORVG_VERSION_STRING "1.0.1"
