/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#define PROJECT_VERSION "${PROJECT_VERSION}"
#define PROJECT_SLUG "${PROJECT_SLUG}"
#define PROJECT_NAME "${PROJECT_NAME}"
#define PROJECT_ID "${PROJECT_ID}"
#define URL_DOCS "${URL_DOCS}"
#define URL_ISSUES "${URL_ISSUES}"
#define URL_DONATE "${URL_DONATE}"
#define PROJECT_DESCRIPTION "${PROJECT_DESCRIPTION}"
#define PROJECT_HOMEPAGE_URL "${PROJECT_HOMEPAGE_URL}"
