/* Copyright JS Foundation and other contributors, http://js.foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* This file is automatically generated by the gen-strings.py script
 * from lit-magic-strings.ini. Do not edit! */

LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING__EMPTY, "")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SPACE_CHAR, " ")
#if JERRY_MODULE_SYSTEM
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASTERISK_CHAR, "*")
#endif /* JERRY_MODULE_SYSTEM */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_COMMA_CHAR, ",")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_E_U, "E")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LEFT_SQUARE_CHAR, "[")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RIGHT_SQUARE_CHAR, "]")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PI_U, "PI")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_AT, "at")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS, "is")
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_OF, "of")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_OR, "or")
#endif /* JERRY_BUILTIN_ATOMICS */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LN2_U, "LN2")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MAP_UL, "Map")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NAN, "NaN")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UL, "Set")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UTC_U, "UTC")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ABS, "abs")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_ATOMICS || JERRY_BUILTIN_CONTAINER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ADD, "add")
#endif /* JERRY_BUILTIN_ATOMICS \
|| JERRY_BUILTIN_CONTAINER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ALL, "all")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_AND, "and")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ANY, "any")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_COS, "cos")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EXP, "exp")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FOR, "for")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET, "get")
#if JERRY_BUILTIN_CONTAINER || JERRY_BUILTIN_PROXY || JERRY_BUILTIN_REFLECT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_HAS, "has")
#endif /* JERRY_BUILTIN_CONTAINER \
|| JERRY_BUILTIN_PROXY            \
|| JERRY_BUILTIN_REFLECT */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOG, "log")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MAP, "map")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MAX, "max")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MIN, "min")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NOW, "now")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_ARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_POP, "pop")
#endif /* JERRY_BUILTIN_ARRAY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_POW, "pow")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RAW, "raw")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET, "set")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SIN, "sin")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_SUB, "sub")
#endif /* JERRY_BUILTIN_ATOMICS */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TAN, "tan")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_XOR, "xor")
#endif /* JERRY_BUILTIN_ATOMICS */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EMPTY_NON_CAPTURE_GROUP, "(?:)")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DATE_UL, "Date")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LN10_U, "LN10")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MATH_UL, "Math")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NULL_UL, "Null")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ACOS, "acos")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASIN, "asin")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATAN, "atan")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BIND, "bind")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CALL, "call")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CBRT, "cbrt")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CEIL, "ceil")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_COSH, "cosh")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DONE, "done")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EVAL, "eval")
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EXEC, "exec")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FILL, "fill")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FIND, "find")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FLAT, "flat")
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FROM, "from")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IMUL, "imul")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_JOIN, "join")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_KEYS, "keys")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_LOAD, "load")
#endif /* JERRY_BUILTIN_ATOMICS */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOG2, "log2")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NAME, "name")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NEXT, "next")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NULL, "null")
#if JERRY_BUILTIN_ARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PUSH, "push")
#endif /* JERRY_BUILTIN_ARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RACE, "race")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SEAL, "seal")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SIGN, "sign")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SINH, "sinh")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_CONTAINER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SIZE, "size")
#endif /* JERRY_BUILTIN_CONTAINER */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SOME, "some")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SORT, "sort")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SQRT, "sqrt")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TANH, "tanh")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TEST, "test")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_THEN, "then")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRIM, "trim")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRUE, "true")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_WAIT, "wait")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ARRAY_UL, "Array")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ERROR_UL, "Error")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOG2E_U, "LOG2E")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_PROXY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PROXY_UL, "Proxy")
#endif /* JERRY_BUILTIN_PROXY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SQRT2_U, "SQRT2")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ACOSH, "acosh")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_APPLY, "apply")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASINH, "asinh")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATAN2, "atan2")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATANH, "atanh")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CATCH, "catch")
#if JERRY_BUILTIN_CONTAINER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CLEAR, "clear")
#endif /* JERRY_BUILTIN_CONTAINER */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CLZ32, "clz32")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_WEAKREF
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DEREF, "deref")
#endif /* JERRY_BUILTIN_WEAKREF */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EVERY, "every")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EXPM1, "expm1")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FALSE, "false")
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FLAGS, "flags")
#endif /* JERRY_BUILTIN_REGEXP */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FLOOR, "floor")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_HYPOT, "hypot")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INDEX, "index")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INPUT, "input")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_NAN, "isNaN")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOG10, "log10")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOG1P, "log1p")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MATCH, "match")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PARSE, "parse")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_PROXY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PROXY, "proxy")
#endif /* JERRY_BUILTIN_PROXY */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ROUND, "round")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_ARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SHIFT, "shift")
#endif /* JERRY_BUILTIN_ARRAY */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_SHAREDARRAYBUFFER || JERRY_BUILTIN_STRING || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SLICE, "slice")
#endif /* JERRY_BUILTIN_ARRAY      \
|| JERRY_BUILTIN_SHAREDARRAYBUFFER \
|| JERRY_BUILTIN_STRING            \
|| JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SPLIT, "split")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_STORE, "store")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_THROW, "throw")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRUNC, "trunc")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_VALUE, "value")
#if JERRY_PARSER && JERRY_SOURCE_NAME
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SOURCE_NAME_EVAL, "<eval>")
#endif /* JERRY_PARSER && JERRY_SOURCE_NAME */
#if JERRY_BUILTIN_BIGINT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BIGINT_UL, "BigInt")
#endif /* JERRY_BUILTIN_BIGINT */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOG10E_U, "LOG10E")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_MODULE_SYSTEM
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MODULE_UL, "Module")
#endif /* JERRY_MODULE_SYSTEM */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NUMBER_UL, "Number")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_OBJECT_UL, "Object")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REGEXP_UL, "RegExp")
#if JERRY_PARSER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SCRIPT_UL, "Script")
#endif /* JERRY_PARSER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_STRING_UL, "String")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SYMBOL_UL, "Symbol")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASSIGN, "assign")
#if JERRY_BUILTIN_BIGINT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BIGINT, "bigint")
#endif /* JERRY_BUILTIN_BIGINT */
#if JERRY_BUILTIN_DATAVIEW || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BUFFER, "buffer")
#endif /* JERRY_BUILTIN_DATAVIEW \
|| JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CALLEE, "callee")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CALLER, "caller")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CHAR_AT_UL, "charAt")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CONCAT, "concat")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CREATE, "create")
#if JERRY_BUILTIN_CONTAINER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DELETE, "delete")
#endif /* JERRY_BUILTIN_CONTAINER */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DOTALL, "dotAll")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ERRORS_UL, "errors")
#if JERRY_BUILTIN_ANNEXB
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ESCAPE, "escape")
#endif /* JERRY_BUILTIN_ANNEXB */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FILTER, "filter")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FREEZE, "freeze")
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FROUND, "fround")
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_DAY_UL, "getDay")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GLOBAL, "global")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_HAS_OWN_UL, "hasOwn")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_VIEW_UL, "isView")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_KEY_FOR, "keyFor")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LENGTH, "length")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_NOTIFY, "notify")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NUMBER, "number")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_OBJECT, "object")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PAD_END, "padEnd")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RANDOM, "random")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REASON, "reason")
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REDUCE, "reduce")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REJECT, "reject")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REPEAT, "repeat")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RETURN, "return")
#if JERRY_BUILTIN_PROXY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REVOKE, "revoke")
#endif /* JERRY_BUILTIN_PROXY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SEARCH, "search")
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SOURCE, "source")
#endif /* JERRY_BUILTIN_REGEXP */
#if JERRY_BUILTIN_ARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SPLICE, "splice")
#endif /* JERRY_BUILTIN_ARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_STATUS, "status")
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_STICKY, "sticky")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_STRING, "string")
#if JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SUBSTR, "substr")
#endif /* JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SYMBOL, "symbol")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_JSON_UL, "toJSON")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_VALUES, "values")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_U, "Atomics")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BOOLEAN_UL, "Boolean")
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EPSILON_U, "EPSILON")
#endif /* JERRY_BUILTIN_NUMBER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PROMISE_UL, "Promise")
#if JERRY_BUILTIN_REFLECT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REFLECT_UL, "Reflect")
#endif /* JERRY_BUILTIN_REFLECT */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SQRT1_2_U, "SQRT1_2")
#endif /* JERRY_BUILTIN_MATH */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SYMBOL_DOT_UL, "Symbol.")
#if JERRY_BUILTIN_CONTAINER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_WEAKMAP_UL, "WeakMap")
#endif /* JERRY_BUILTIN_CONTAINER */
#if JERRY_BUILTIN_WEAKREF
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_WEAKREF_UL, "WeakRef")
#endif /* JERRY_BUILTIN_WEAKREF */
#if JERRY_BUILTIN_CONTAINER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_WEAKSET_UL, "WeakSet")
#endif /* JERRY_BUILTIN_CONTAINER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BOOLEAN, "boolean")
#if JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_COMPILE, "compile")
#endif /* JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DEFAULT, "default")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ENTRIES, "entries")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FINALLY, "finally")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FLATMAP, "flatMap")
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_CONTAINER || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FOR_EACH_UL, "forEach")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_CONTAINER    \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_DATE_UL, "getDate")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_INT8_UL, "getInt8")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_TIME_UL, "getTime")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_YEAR_UL, "getYear")
#endif /* JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_STRING || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INDEX_OF_UL, "indexOf")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_STRING       \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_ARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_ARRAY_UL, "isArray")
#endif /* JERRY_BUILTIN_ARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MESSAGE, "message")
#if JERRY_BUILTIN_PROXY || JERRY_BUILTIN_REFLECT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_OWN_KEYS_UL, "ownKeys")
#endif /* JERRY_BUILTIN_PROXY \
|| JERRY_BUILTIN_REFLECT */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REPLACE, "replace")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RESOLVE, "resolve")
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REVERSE, "reverse")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_DATE_UL, "setDate")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_INT8_UL, "setInt8")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_TIME_UL, "setTime")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_YEAR_UL, "setYear")
#endif /* JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SPECIES, "species")
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_FIXED_UL, "toFixed")
#endif /* JERRY_BUILTIN_NUMBER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRIM_END, "trimEnd")
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UNICODE, "unicode")
#endif /* JERRY_BUILTIN_REGEXP */
#if JERRY_BUILTIN_ARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UNSHIFT, "unshift")
#endif /* JERRY_BUILTIN_ARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_VALUE_OF_UL, "valueOf")
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DATAVIEW_UL, "DataView")
#endif /* JERRY_BUILTIN_DATAVIEW */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FUNCTION_UL, "Function")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INFINITY_UL, "Infinity")
#if JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_URI_ERROR_UL, "URIError")
#endif /* JERRY_BUILTIN_ERRORS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_OBJECT_TO_STRING_UL, "[object ")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ENDS_WITH, "endsWith")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_EXCHANGE, "exchange")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FUNCTION, "function")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_HOURS_UL, "getHours")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_INT16_UL, "getInt16")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_INT32_UL, "getInt32")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_MONTH_UL, "getMonth")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UINT8_UL, "getUint8")
#endif /* JERRY_BUILTIN_DATAVIEW */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INCLUDES, "includes")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_FINITE, "isFinite")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_FROZEN_UL, "isFrozen")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_SEALED_UL, "isSealed")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ITERATOR, "iterator")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MATCH_ALL, "matchAll")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PAD_START, "padStart")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PARSE_INT, "parseInt")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REJECTED, "rejected")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_HOURS_UL, "setHours")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_INT16_UL, "setInt16")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_INT32_UL, "setInt32")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_MONTH_UL, "setMonth")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UINT8_UL, "setUint8")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SUBARRAY, "subarray")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_STRING_UL, "toString")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRIM_LEFT, "trimLeft")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_ANNEXB
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UNESCAPE, "unescape")
#endif /* JERRY_BUILTIN_ANNEXB */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_WRITABLE, "writable")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NEGATIVE_INFINITY_UL, "-Infinity")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ARGUMENTS_UL, "Arguments")
#if JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_EVAL_ERROR_UL, "EvalError")
#endif /* JERRY_BUILTIN_ERRORS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GENERATOR_UL, "Generator")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INT8_ARRAY_UL, "Int8Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MAX_VALUE_U, "MAX_VALUE")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MIN_VALUE_U, "MIN_VALUE")
#endif /* JERRY_BUILTIN_NUMBER */
#if JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TYPE_ERROR_UL, "TypeError")
#endif /* JERRY_BUILTIN_ERRORS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UNDEFINED_UL, "Undefined")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING__PROTO__, "__proto__")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ANONYMOUS, "anonymous")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ARGUMENTS, "arguments")
#if JERRY_BUILTIN_PROXY || JERRY_BUILTIN_REFLECT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CONSTRUCT, "construct")
#endif /* JERRY_BUILTIN_PROXY \
|| JERRY_BUILTIN_REFLECT */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DECODE_URI, "decodeURI")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ENCODE_URI, "encodeURI")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FIND_INDEX, "findIndex")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FULFILLED, "fulfilled")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_DAY_UL, "getUTCDay")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UINT16_UL, "getUint16")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UINT32_UL, "getUint32")
#endif /* JERRY_BUILTIN_DATAVIEW */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_INTEGER, "isInteger")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LASTINDEX_UL, "lastIndex")
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MULTILINE, "multiline")
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PROTOTYPE, "prototype")
#if JERRY_BUILTIN_PROXY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REVOCABLE, "revocable")
#endif /* JERRY_BUILTIN_PROXY */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UINT16_UL, "setUint16")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UINT32_UL, "setUint32")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SUBSTRING, "substring")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRIM_RIGHT, "trimRight")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TRIM_START, "trimStart")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UNDEFINED, "undefined")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INT16_ARRAY_UL, "Int16Array")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INT32_ARRAY_UL, "Int32Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_RANGE_ERROR_UL, "RangeError")
#endif /* JERRY_BUILTIN_ERRORS */
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TYPED_ARRAY_UL, "TypedArray")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UINT8_ARRAY_UL, "Uint8Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ALLSETTLED, "allSettled")
#if JERRY_BUILTIN_DATAVIEW || JERRY_BUILTIN_SHAREDARRAYBUFFER || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BYTE_LENGTH_UL, "byteLength")
#endif /* JERRY_BUILTIN_DATAVIEW   \
|| JERRY_BUILTIN_SHAREDARRAYBUFFER \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_DATAVIEW || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BYTE_OFFSET_UL, "byteOffset")
#endif /* JERRY_BUILTIN_DATAVIEW \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CHAR_CODE_AT_UL, "charCodeAt")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_COPY_WITHIN, "copyWithin")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ENUMERABLE, "enumerable")
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_FLOAT_32_UL, "getFloat32")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATAVIEW && JERRY_NUMBER_TYPE_FLOAT64
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_FLOAT_64_UL, "getFloat64")
#endif /* JERRY_BUILTIN_DATAVIEW && JERRY_NUMBER_TYPE_FLOAT64 */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_MINUTES_UL, "getMinutes")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_SECONDS_UL, "getSeconds")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_DATE_UL, "getUTCDate")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_GLOBAL_THIS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GLOBAL_THIS_UL, "globalThis")
#endif /* JERRY_BUILTIN_GLOBAL_THIS */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IGNORECASE_UL, "ignoreCase")
#endif /* JERRY_BUILTIN_REGEXP */
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_ISLOCKFREE, "isLockFree")
#endif /* JERRY_BUILTIN_ATOMICS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PARSE_FLOAT, "parseFloat")
#if JERRY_BUILTIN_REGEXP && JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REPLACE_ALL, "replaceAll")
#endif /* JERRY_BUILTIN_REGEXP && JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_FLOAT_32_UL, "setFloat32")
#endif /* JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATAVIEW && JERRY_NUMBER_TYPE_FLOAT64
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_FLOAT_64_UL, "setFloat64")
#endif /* JERRY_BUILTIN_DATAVIEW && JERRY_NUMBER_TYPE_FLOAT64 */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_MINUTES_UL, "setMinutes")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_SECONDS_UL, "setSeconds")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_DATE_UL, "setUTCDate")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_STARTS_WITH, "startsWith")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SOURCE_NAME_ANON, "<anonymous>")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ARRAY_BUFFER_UL, "ArrayBuffer")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SYNTAX_ERROR_UL, "SyntaxError")
#endif /* JERRY_BUILTIN_ERRORS */
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UINT16_ARRAY_UL, "Uint16Array")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UINT32_ARRAY_UL, "Uint32Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CODE_POINT_AT, "codePointAt")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CONSTRUCTOR, "constructor")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DESCRIPTION, "description")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_OBJECT_FROM_ENTRIES, "fromEntries")
#if JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_BIGINT64, "getBigInt64")
#endif /* JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_FULL_YEAR_UL, "getFullYear")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_HOURS_UL, "getUTCHours")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_MONTH_UL, "getUTCMonth")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_HAS_INSTANCE, "hasInstance")
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_STRING || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LAST_INDEX_OF_UL, "lastIndexOf")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_STRING       \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_ARRAY || JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REDUCE_RIGHT_UL, "reduceRight")
#endif /* JERRY_BUILTIN_ARRAY \
|| JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_BIGINT64, "setBigInt64")
#endif /* JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_FULL_YEAR_UL, "setFullYear")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_HOURS_UL, "setUTCHours")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_MONTH_UL, "setUTCMonth")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_GMT_STRING_UL, "toGMTString")
#endif /* JERRY_BUILTIN_ANNEXB && JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_ISO_STRING_UL, "toISOString")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_LOWER_CASE_UL, "toLowerCase")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_PRECISION_UL, "toPrecision")
#endif /* JERRY_BUILTIN_NUMBER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_PRIMITIVE, "toPrimitive")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_STRING_TAG, "toStringTag")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_UTC_STRING_UL, "toUTCString")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_UPPER_CASE_UL, "toUpperCase")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UNSCOPABLES, "unscopables")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FLOAT32_ARRAY_UL, "Float32Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_TYPEDARRAY && JERRY_NUMBER_TYPE_FLOAT64
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FLOAT64_ARRAY_UL, "Float64Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY && JERRY_NUMBER_TYPE_FLOAT64 */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_INVALID_DATE_UL, "Invalid Date")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MAP_ITERATOR_UL, "Map Iterator")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_ITERATOR_UL, "Set Iterator")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_CONFIGURABLE, "configurable")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FROM_CHAR_CODE_UL, "fromCharCode")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_BIGUINT64, "getBigUint64")
#endif /* JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_EXTENSIBLE, "isExtensible")
#if JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_BIGUINT64, "setBigUint64")
#endif /* JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_DATAVIEW */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_DATE_STRING_UL, "toDateString")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_TIME_STRING_UL, "toTimeString")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASYNC_FUNCTION_UL, "AsyncFunction")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BIGINT64_ARRAY_UL, "BigInt64Array")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASYNC_ITERATOR, "asyncIterator")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FROM_CODE_POINT_UL, "fromCodePoint")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_MINUTES_UL, "getUTCMinutes")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_SECONDS_UL, "getUTCSeconds")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_PROTOTYPE_OF_UL, "isPrototypeOf")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_SAFE_INTEGER, "isSafeInteger")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOCALE_COMPARE_UL, "localeCompare")
#endif /* JERRY_BUILTIN_STRING */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_MINUTES_UL, "setUTCMinutes")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_SECONDS_UL, "setUTCSeconds")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_EXPONENTIAL_UL, "toExponential")
#endif /* JERRY_BUILTIN_NUMBER */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_AGGREGATE_ERROR_UL, "AggregateError")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ARRAY_ITERATOR_UL, "Array Iterator")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASYNC_GENERATOR_UL, "AsyncGenerator")
#if JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BIGUINT64_ARRAY_UL, "BigUint64Array")
#endif /* JERRY_BUILTIN_BIGINT && JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REFERENCE_ERROR_UL, "ReferenceError")
#endif /* JERRY_BUILTIN_ERRORS */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DEFINE_PROPERTY_UL, "defineProperty")
#if JERRY_BUILTIN_PROXY || JERRY_BUILTIN_REFLECT
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DELETE_PROPERTY_UL, "deleteProperty")
#endif /* JERRY_BUILTIN_PROXY \
|| JERRY_BUILTIN_REFLECT */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_PROTOTYPE_OF_UL, "getPrototypeOf")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_FULL_YEAR_UL, "getUTCFullYear")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_HAS_OWN_PROPERTY_UL, "hasOwnProperty")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_PROTOTYPE_OF_UL, "setPrototypeOf")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_FULL_YEAR_UL, "setUTCFullYear")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_LOCALE_STRING_UL, "toLocaleString")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_STRING_ITERATOR_UL, "String Iterator")
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ATOMICS_COMPAREEXCHANGE, "compareExchange")
#endif /* JERRY_BUILTIN_ATOMICS */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_MILLISECONDS_UL, "getMilliseconds")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_MILLISECONDS_UL, "setMilliseconds")
#endif /* JERRY_BUILTIN_DATE */
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MAX_SAFE_INTEGER_U, "MAX_SAFE_INTEGER")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_MIN_SAFE_INTEGER_U, "MIN_SAFE_INTEGER")
#endif /* JERRY_BUILTIN_NUMBER */
#if JERRY_BUILTIN_ANNEXB
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DEFINE_GETTER, "__defineGetter__")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DEFINE_SETTER, "__defineSetter__")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOOKUP_GETTER, "__lookupGetter__")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_LOOKUP_SETTER, "__lookupSetter__")
#endif /* JERRY_BUILTIN_ANNEXB */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DEFINE_PROPERTIES_UL, "defineProperties")
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_BYTES_PER_ELEMENT_U, "BYTES_PER_ELEMENT")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GENERATOR_FUNCTION_UL, "GeneratorFunction")
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_NEGATIVE_INFINITY_U, "NEGATIVE_INFINITY")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_POSITIVE_INFINITY_U, "POSITIVE_INFINITY")
#endif /* JERRY_BUILTIN_NUMBER */
#if JERRY_BUILTIN_SHAREDARRAYBUFFER
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SHARED_ARRAY_BUFFER_UL, "SharedArrayBuffer")
#endif /* JERRY_BUILTIN_SHAREDARRAYBUFFER */
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_UINT8_CLAMPED_ARRAY_UL, "Uint8ClampedArray")
#endif /* JERRY_BUILTIN_TYPEDARRAY */
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_TIMEZONE_OFFSET_UL, "getTimezoneOffset")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PREVENT_EXTENSIONS_UL, "preventExtensions")
#if JERRY_BUILTIN_STRING
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_LOCALE_LOWER_CASE_UL, "toLocaleLowerCase")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_LOCALE_UPPER_CASE_UL, "toLocaleUpperCase")
#endif /* JERRY_BUILTIN_STRING */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_DECODE_URI_COMPONENT, "decodeURIComponent")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ENCODE_URI_COMPONENT, "encodeURIComponent")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_UTC_MILLISECONDS_UL, "getUTCMilliseconds")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_IS_CONCAT_SPREADABLE, "isConcatSpreadable")
#if JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_SET_UTC_MILLISECONDS_UL, "setUTCMilliseconds")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_LOCALE_DATE_STRING_UL, "toLocaleDateString")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_TO_LOCALE_TIME_STRING_UL, "toLocaleTimeString")
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_OWN_PROPERTY_NAMES_UL, "getOwnPropertyNames")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_PROPERTY_IS_ENUMERABLE_UL, "propertyIsEnumerable")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_OWN_PROPERTY_SYMBOLS_UL, "getOwnPropertySymbols")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_ASYNC_GENERATOR_FUNCTION_UL, "AsyncGeneratorFunction")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_REGEXP_STRING_ITERATOR_UL, "RegExp String Iterator")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_OWN_PROPERTY_DESCRIPTOR_UL, "getOwnPropertyDescriptor")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_GET_OWN_PROPERTY_DESCRIPTORS_UL, "getOwnPropertyDescriptors")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FUNCTION_TO_STRING_NATIVE, "function () { [native code] }")
LIT_MAGIC_STRING_DEF (LIT_MAGIC_STRING_FUNCTION_TO_STRING_ECMA, "function () { /* ecmascript */ }")

LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (0, LIT_MAGIC_STRING__EMPTY)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (1, LIT_MAGIC_STRING_SPACE_CHAR)
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (2, LIT_MAGIC_STRING_PI_U)
#else /* !(JERRY_BUILTIN_MATH) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (2, LIT_MAGIC_STRING_AT)
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (3, LIT_MAGIC_STRING_LN2_U)
#else /* !(JERRY_BUILTIN_MATH) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (3, LIT_MAGIC_STRING_MAP_UL)
#endif /* JERRY_BUILTIN_MATH */
#if JERRY_BUILTIN_REGEXP
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (4, LIT_MAGIC_STRING_EMPTY_NON_CAPTURE_GROUP)
#else /* !(JERRY_BUILTIN_REGEXP) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (4, LIT_MAGIC_STRING_DATE_UL)
#endif /* JERRY_BUILTIN_REGEXP */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (5, LIT_MAGIC_STRING_ARRAY_UL)
#if JERRY_PARSER && JERRY_SOURCE_NAME
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (6, LIT_MAGIC_STRING_SOURCE_NAME_EVAL)
#elif JERRY_BUILTIN_BIGINT
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (6, LIT_MAGIC_STRING_BIGINT_UL)
#elif JERRY_BUILTIN_MATH
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (6, LIT_MAGIC_STRING_LOG10E_U)
#elif JERRY_MODULE_SYSTEM
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (6, LIT_MAGIC_STRING_MODULE_UL)
#else /* !(JERRY_MODULE_SYSTEM) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (6, LIT_MAGIC_STRING_NUMBER_UL)
#endif /* JERRY_MODULE_SYSTEM */
#if JERRY_BUILTIN_ATOMICS
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (7, LIT_MAGIC_STRING_ATOMICS_U)
#else /* !(JERRY_BUILTIN_ATOMICS) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (7, LIT_MAGIC_STRING_BOOLEAN_UL)
#endif /* JERRY_BUILTIN_ATOMICS */
#if JERRY_BUILTIN_DATAVIEW
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (8, LIT_MAGIC_STRING_DATAVIEW_UL)
#else /* !(JERRY_BUILTIN_DATAVIEW) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (8, LIT_MAGIC_STRING_FUNCTION_UL)
#endif /* JERRY_BUILTIN_DATAVIEW */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (9, LIT_MAGIC_STRING_NEGATIVE_INFINITY_UL)
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (10, LIT_MAGIC_STRING_INT16_ARRAY_UL)
#elif JERRY_BUILTIN_ERRORS
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (10, LIT_MAGIC_STRING_RANGE_ERROR_UL)
#elif JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (10, LIT_MAGIC_STRING_TYPED_ARRAY_UL)
#else /* !(JERRY_BUILTIN_TYPEDARRAY) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (10, LIT_MAGIC_STRING_ALLSETTLED)
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (11, LIT_MAGIC_STRING_SOURCE_NAME_ANON)
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (12, LIT_MAGIC_STRING_FLOAT32_ARRAY_UL)
#elif JERRY_BUILTIN_TYPEDARRAY && JERRY_NUMBER_TYPE_FLOAT64
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (12, LIT_MAGIC_STRING_FLOAT64_ARRAY_UL)
#elif JERRY_BUILTIN_DATE
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (12, LIT_MAGIC_STRING_INVALID_DATE_UL)
#else /* !(JERRY_BUILTIN_DATE) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (12, LIT_MAGIC_STRING_MAP_ITERATOR_UL)
#endif /* JERRY_BUILTIN_DATE */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (13, LIT_MAGIC_STRING_ASYNC_FUNCTION_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (14, LIT_MAGIC_STRING_AGGREGATE_ERROR_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (15, LIT_MAGIC_STRING_STRING_ITERATOR_UL)
#if JERRY_BUILTIN_NUMBER
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (16, LIT_MAGIC_STRING_MAX_SAFE_INTEGER_U)
#elif JERRY_BUILTIN_ANNEXB
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (16, LIT_MAGIC_STRING_DEFINE_GETTER)
#else /* !(JERRY_BUILTIN_ANNEXB) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (16, LIT_MAGIC_STRING_DEFINE_PROPERTIES_UL)
#endif /* JERRY_BUILTIN_ANNEXB */
#if JERRY_BUILTIN_TYPEDARRAY
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (17, LIT_MAGIC_STRING_BYTES_PER_ELEMENT_U)
#else /* !(JERRY_BUILTIN_TYPEDARRAY) */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (17, LIT_MAGIC_STRING_GENERATOR_FUNCTION_UL)
#endif /* JERRY_BUILTIN_TYPEDARRAY */
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (18, LIT_MAGIC_STRING_DECODE_URI_COMPONENT)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (19, LIT_MAGIC_STRING_GET_OWN_PROPERTY_NAMES_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (20, LIT_MAGIC_STRING_PROPERTY_IS_ENUMERABLE_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (21, LIT_MAGIC_STRING_GET_OWN_PROPERTY_SYMBOLS_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (22, LIT_MAGIC_STRING_ASYNC_GENERATOR_FUNCTION_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (23, LIT_MAGIC_STRING_GET_OWN_PROPERTY_DESCRIPTOR_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (24, LIT_MAGIC_STRING_GET_OWN_PROPERTY_DESCRIPTOR_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (25, LIT_MAGIC_STRING_GET_OWN_PROPERTY_DESCRIPTORS_UL)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (26, LIT_MAGIC_STRING_FUNCTION_TO_STRING_NATIVE)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (27, LIT_MAGIC_STRING_FUNCTION_TO_STRING_NATIVE)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (28, LIT_MAGIC_STRING_FUNCTION_TO_STRING_NATIVE)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (29, LIT_MAGIC_STRING_FUNCTION_TO_STRING_NATIVE)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (30, LIT_MAGIC_STRING_FUNCTION_TO_STRING_ECMA)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (31, LIT_MAGIC_STRING_FUNCTION_TO_STRING_ECMA)
LIT_MAGIC_STRING_FIRST_STRING_WITH_SIZE (32, LIT_MAGIC_STRING_FUNCTION_TO_STRING_ECMA)
