/* Copyright JS Foundation and other contributors, http://js.foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* This file is automatically generated by the gen-unicode.py script
 * from UnicodeData.txt and SpecialCasing.txt files. Do not edit! */

/* Contains start points of character case ranges (these are bidirectional conversions). */
static const uint32_t lit_unicode_character_case_ranges_sup[] JERRY_ATTR_CONST_DATA = {
  0x010400, 0x010428, 0x0104b0, 0x0104d8, 0x010c80, 0x010cc0, 0x0118a0, 0x0118c0, 0x016e40, 0x016e60, 0x01e900, 0x01e922
};

/* Interval lengths of start points in `character_case_ranges` table. */
static const uint16_t lit_unicode_character_case_range_lengths_sup[] JERRY_ATTR_CONST_DATA = { 0x000028, 0x000024,
                                                                                               0x000033, 0x000020,
                                                                                               0x000020, 0x000022 };
