/* Copyright JS Foundation and other contributors, http://js.foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* This file is automatically generated by the gen-unicode.py script
 * from the CaseFolding.txt file. Do not edit! */

/**
 * Character interval starting points for folding_skip_to_lower.
 */
static const uint16_t lit_unicode_folding_skip_to_lower_interval_starts[] JERRY_ATTR_CONST_DATA = { 0x13a0,
                                                                                                    0x13f8,
                                                                                                    0xab70 };

/**
 * Character interval lengths for folding_skip_to_lower.
 */
static const uint8_t lit_unicode_folding_skip_to_lower_interval_lengths[] JERRY_ATTR_CONST_DATA = { 0x0055,
                                                                                                    0x0005,
                                                                                                    0x004f };

/**
 * Non-interval characters for folding_skip_to_lower.
 */
static const uint16_t lit_unicode_folding_skip_to_lower_chars[] JERRY_ATTR_CONST_DATA = { 0x0130 };

/**
 * Character interval starting points for folding_to_upper.
 */
static const uint16_t lit_unicode_folding_to_upper_interval_starts[] JERRY_ATTR_CONST_DATA = { 0x03d0, 0x03d5, 0x03f0,
                                                                                               0x13f8, 0x1c80, 0xab70 };

/**
 * Character interval lengths for folding_to_upper.
 */
static const uint8_t lit_unicode_folding_to_upper_interval_lengths[] JERRY_ATTR_CONST_DATA = { 0x0001, 0x0001, 0x0001,
                                                                                               0x0005, 0x0008, 0x004f };

/**
 * Non-interval characters for folding_to_upper.
 */
static const uint16_t lit_unicode_folding_to_upper_chars[] JERRY_ATTR_CONST_DATA = { 0x00b5, 0x017f, 0x0345, 0x03c2,
                                                                                     0x03f5, 0x1e9b, 0x1fbe };
