/*
 * SPDX-FileCopyrightText: 2019-2026 Mattia Basaglia <dev@dragon.best>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include "glaxnimate/io/base.hpp"


namespace glaxnimate::io::rive {

class RiveHtmlFormat : public ImportExport
{
    Q_OBJECT

public:
    QString slug() const override { return "rive_html"; }
    QString name() const override { return i18n("RIVE HTML Preview"); }
    QStringList extensions(Direction) const override { return {"html", "htm"}; }
    bool can_save() const override { return true; }
    bool can_open() const override { return false; }

private:
    bool on_save(QIODevice& file, const QString& filename,
                 model::Composition* comp, const QVariantMap& setting_values) override;
};

} // namespace glaxnimate::io::rive
